#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu May  2 17:56:49 2019

Class for handling data saving.

For now, this is how it works: you choose a directory and you create a Storage object with the directory as input. 
Then, the method create_data_file creates a data file in dump, to which append_data_to_current_file can be used to load data 
into it. That lies in the dump, and can me made permanent with permanent_save_current_data. 

@author: martindrech
"""

import numpy as np
import datetime
import os
import shutil

class Storage(object):
    """
    Class to handle data saving. 
    """
    def __init__(self):
        """
        By default, directory is set to the current directory. 
        """
        self.directory = os.getcwd()
    
    def set_directory(self, directory):
        self.directory = directory
        
    def create_data_file(self, *args):
        """
        Creates a file to store data. Name is set automatically. 
        *args should be strings, with the column titles of the data to be saved. 
        """
        saving_directory = "/".join([self.directory, self.get_date_string()])
        if not os.path.exists(saving_directory):
            os.makedirs(saving_directory)
            os.makedirs(saving_directory + '/dump')

        s_num = self.get_saving_number()
       
        # data file
        filename = "/".join([saving_directory, 'dump', 's' + str(s_num).zfill(2)]) + '.temp'
        header = " ".join([s for s in args]+['\n'])
        with open(filename, 'wb') as f:
            f.write(header.encode('utf-8'))
        #meta data
        filename = "/".join([saving_directory, 'dump', 'metadata' + str(s_num).zfill(2)]) + '.txt'
        header = 'Meta data file'
        with open(filename, 'wb') as f:
            f.write(header.encode('utf-8'))
        
        self.delete_temporary()
        
        
    def append_data_to_current_file(self, row):
        """
        This will append a row to the last data file created. 
        row should be a list with the values to append. 
        """
        with open(self.get_current_data_filename(), 'ab') as f:
            #data = np.column_stack(args)
            np.savetxt(f, [row])
            f.flush()

    def append_metadata_to_current_file(self, *args):
        """
        This appends metadata to the last metadata file created. 
        *args should be dictionaries, each of them will be stored. 
        """
        for arg in args:
            for key, value in arg.items():
                with open(self.get_current_metadata_filename(), 'a') as f:
                    f.write('\n' + str(key) + ' ' + str(value))
                    f.flush()
            with open(self.get_current_metadata_filename(), 'a') as f:
                    f.write('\n')
                    f.flush()
        pass
    
    def get_date_string(self):
        y = datetime.datetime.now().year
        m = datetime.datetime.now().month
        d = datetime.datetime.now().day
        s = str(y) + str(m).zfill(2) + str(d).zfill(2)
        return s
        
    def get_saving_number(self, permanent_saving = False):
        
        savings_number_list = []
        if permanent_saving:
            suffix = '.dat'
            saving_directory = "/".join([self.directory, self.get_date_string()])
        else:
            suffix = '.temp'
            saving_directory = "/".join([self.directory, self.get_date_string(), 'dump'])
        
        for s in [dat_file for dat_file in os.listdir(saving_directory) if dat_file.endswith(suffix)]:
            savings_number_list.append(int(s[1:].split('.')[0]))
        
        if len(savings_number_list) == 0:
            return 1      
                      
        return max(savings_number_list)+1
    
    def delete_temporary(self):
        """
        Deletes old temporary files. 
        """
        savings_number_list = []
        suffix = '.temp'
        saving_directory = "/".join([self.directory, self.get_date_string(), 'dump'])
        
        for s in [dat_file for dat_file in os.listdir(saving_directory) if dat_file.endswith(suffix)]:
            savings_number_list.append(int(s[1:].split('.')[0]))
            
        if len(savings_number_list) > 3: #removes the last of the last 10 files
            number_to_remove = min(savings_number_list)
            # data file
            file_to_remove = "/".join([saving_directory, 's' + str(number_to_remove).zfill(2)])+'.temp'
            os.remove(file_to_remove)            
            #meta data
            file_to_remove = "/".join([saving_directory, 'metadata' + str(number_to_remove).zfill(2)])+'.txt'
            os.remove(file_to_remove)            
            
    def get_current_data_filename(self):
        s_num = self.get_saving_number()-1
        saving_directory = "/".join([self.directory, self.get_date_string()])
        filename = "/".join([saving_directory, 'dump', 's' + str(s_num).zfill(2)]) + '.temp'
        return filename

    def get_current_metadata_filename(self):
        s_num = self.get_saving_number()-1
        saving_directory = "/".join([self.directory, self.get_date_string()])
        filename = "/".join([saving_directory, 'dump', 'metadata' + str(s_num).zfill(2)]) + '.txt'
        return filename


    def permanent_save_current_data(self):
        saving_number = self.get_saving_number(permanent_saving = True)
        old_datafile = self.get_current_data_filename()
        old_metadatafile = self.get_current_metadata_filename()
        saving_directory = "/".join([self.directory, self.get_date_string()])
        # data file
        newfilename = 's' + str(saving_number).zfill(2) + '.dat'
        shutil.move(old_datafile, "/".join([saving_directory, newfilename]) )
        # metadata file
        newfilename = 'metadata' + str(saving_number ).zfill(2) + '.txt'
        shutil.move(old_metadatafile, "/".join([saving_directory, newfilename]) )
        
        
        

#%%
if __name__ == '__main__':
    storage = Storage()
    storage.create_data_file('col1', 'col2')
    storage.create_data_file('col1', 'col2')
    storage.create_data_file('col1', 'col2')
    storage.create_data_file('col1', 'col2')
    storage.create_data_file('col1', 'col2', 'col3')
    d1 = {'nombre': 'martin', 'apellido': 'drechsler'}           
    d2 = {'nombre': 'main', 'edad': 5}
    storage.append_metadata_to_current_file(d1, d2) 
    
    
    