from pyqtgraph import dockarea
from pyqtgraph.Qt import QtGui, QtCore
from PyQt5.QtWidgets import QPushButton, QProgressBar, QLabel
import pyqtgraph as pg
import numpy as np
import resources.scanFunctions as scanFunctions
from resources.messages import show_warning



class CustomSpinBox(pg.SpinBox):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.opts['suffix'] == 'V': # fixing a context menu with some scales 
            self.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
            self.customContextMenuRequested.connect(self._on_context_menu)
            self.popMenu = QtGui.QMenu(self)
            self.act_2_step = QtGui.QAction('2 mV step', self)
            self.act_5_step = QtGui.QAction('5 mV step', self)
            self.act_10_step = QtGui.QAction('10 mV step', self)
            self.act_100_step = QtGui.QAction('100 mV step', self)

            self.popMenu.addAction(self.act_2_step)
            self.popMenu.addAction(self.act_5_step)   
            self.popMenu.addAction(self.act_10_step)
            self.popMenu.addAction(self.act_100_step)

            self.act_2_step.triggered.connect(lambda: self.setSingleStep(0.002))
            self.act_5_step.triggered.connect(lambda: self.setSingleStep(0.005))
            self.act_10_step.triggered.connect(lambda: self.setSingleStep(0.01))
            self.act_100_step.triggered.connect(lambda: self.setSingleStep(0.1))

    def _on_context_menu(self, point):
        # show context menu
        self.popMenu.exec_(self.mapToGlobal(point)) 

class LaserControlLayout(pg.LayoutWidget):
    """
    A class for creating a layout with spinboxes and checkboxes to control a laser. 
    """
    _registry = []
    blinkSignal = QtCore.pyqtSignal(bool, float)

    def __init__(self, name, spinboxes = ['sbA', 'sbB'], checkboxes = ['cbA', 'cbB'], custom_geometry = False, blinkCheckbox = []):
        super().__init__()
        self._registry.append(self)
        self.name = name
        if custom_geometry != True:
            self.layout.addWidget(QtGui.QLabel(self.name), 0, 0)
        

        self.spinboxes = {}
        self.checkboxes = {}

        for i, sb_name in enumerate(spinboxes, start = 1):
            self.spinboxes[sb_name] = CustomSpinBox(siPrefix=True, suffix='V')
            if custom_geometry != True:
                self.addWidget(QtGui.QLabel(sb_name), i, 0)
                self.addWidget(self.spinboxes[sb_name], i, 1)

        for i, cb_name in enumerate(checkboxes, start = 0):
            self.checkboxes[cb_name] = QtGui.QCheckBox()
            if custom_geometry != True:
                row = len(spinboxes) + 1
                self.addWidget(QtGui.QLabel(cb_name), row, 2*i)
                self.addWidget(self.checkboxes[cb_name], row, 2*i + 1)

        self.blinkCheckboxes = {}
        self.blinkSpinboxes = {}
        self.blinkTimers = {}        
        for i, cb_name in enumerate(blinkCheckbox, start=0):
            self.blinkCheckboxes[cb_name] = QtGui.QCheckBox()
            self.blinkCheckboxes[cb_name].stateChanged.connect(lambda: self.blink(cb_name))
            self.blinkSpinboxes[cb_name] = CustomSpinBox(siPrefix=True, suffix='s', value = 1)
            self.blinkSpinboxes[cb_name].setMinimum(0.01)
            self.blinkTimers[cb_name] = QtCore.QTimer()
            self.blinkTimers[cb_name].timeout.connect(self.checkboxes[cb_name].toggle)
            if custom_geometry != True:
                row = 2 + len(self.spinboxes)
                self.addWidget(QtGui.QLabel('blink '+cb_name), row, 0)
                self.addWidget(self.blinkCheckboxes[cb_name], row, 1)
                self.addWidget(self.blinkSpinboxes[cb_name], row, 2)


            
    def freeze(self):
        for sb in self.spinboxes:
            self.spinboxes[sb].setEnabled(False)
        for cb in self.checkboxes:
            self.checkboxes[cb].setEnabled(False)

    def unfreeze(self):
        for sb in self.spinboxes:
            self.spinboxes[sb].setEnabled(True)
        for cb in self.checkboxes:
            self.checkboxes[cb].setEnabled(True)


    def blink(self, cb_name):
        T = self.blinkSpinboxes[cb_name].value()
        if self.blinkCheckboxes[cb_name].isChecked():
            print(' '.join((cb_name, 'blinking with period: ')), T)
            self._initial_state = self.checkboxes[cb_name].isChecked()
            self.blinkTimers[cb_name].start(T*1e3)
        else:
            print(' '.join((cb_name, 'blinking stopped')))
            self.blinkTimers[cb_name].stop()
            self.checkboxes[cb_name].setChecked(self._initial_state)

class LaserScanLayout(pg.LayoutWidget):
    """
    A class for creating a layout for scanning a laser piezo. 
    It is important that the name parameter coincides with the name of the spinbox to scan. 
    """
    _registry = []
    scanActionSignal = QtCore.pyqtSignal(object)

    def __init__(self, name, signal_to_scan_layout):
        super().__init__()
        self._registry.append(self)
        self.name = name
        self.signal_to_scan_layout = signal_to_scan_layout
        self.addWidget(QtGui.QLabel(self.name), 0, 0)
        self.spinboxes = {}
        self.spinboxes['start'] = CustomSpinBox(siPrefix=True, suffix='V')
        self.spinboxes['stop'] = CustomSpinBox(siPrefix=True, suffix='V')
        self.spinboxes['period'] = CustomSpinBox(siPrefix=True, suffix='s', step=1)

        self.spinboxes['period'].setMinimum(0)
        
        self.scan_button = QPushButton('Start scan')
        self.scanpBar = QProgressBar(self)
        self.scanLabelValue = QLabel(self)

        self.addWidget(QtGui.QLabel('start'), 1, 0),
        self.addWidget(self.spinboxes['start'], 1, 1)
        self.addWidget(QtGui.QLabel('stop'), 2, 0) 
        self.addWidget(self.spinboxes['stop'], 2, 1)
        self.addWidget(QtGui.QLabel('period'), 3, 0),
        self.addWidget(self.spinboxes['period'], 3, 1)
        self.addWidget(QtGui.QLabel('Scanning'), 4, 0)
        self.addWidget(self.scan_button, 4, 1)
        
        self.addWidget(self.scanLabelValue, 5, 0)
        self.addWidget(self.scanpBar, 5, 1)
    
        self.scan_button.clicked.connect(self._scan_button_clicked)

    def name(self):
        return self.name
        
    @QtCore.pyqtSlot()
    def _scan_button_clicked(self):

        start = self.spinboxes['start'].value()
        stop = self.spinboxes['stop'].value()
        period = self.spinboxes['period'].value()
        dt = 5.0e-3            
        first_value = float(self.signal_to_scan_layout.spinboxes[self.name].value())            
        scan_array = scanFunctions.create_scan_array(start, stop, period, first_value, dt)
        scan_array_gen = scanFunctions.yield_scan_array(scan_array)
        if first_value > float(stop) or first_value < float(start):
            details = 'Remember remember the fifth of November.' + '\n' + 'Also remember that scan starts from the value of the corresponding spin box'
            show_warning('Scan can not start outside scan range and scan start should be lower than scan stop', details_text= details)
        elif len(scan_array) < 5:
            show_warning('%i steps are too few, you need at least 5 steps' % (len(scan_array)) )
        elif stop==start:
            show_warning('Oh dear, you might be trying to scan from 0 to 0. Plase never do that again.')

        else:
            self.scanActionSignal.emit(scan_array_gen)
            self.scan_button.setText('Stop scan')
            self.signal_to_scan_layout.spinboxes[self.name].setEnabled(False)
            self.spinboxes['start'].setEnabled(False)
            self.spinboxes['stop'].setEnabled(False)
            self.spinboxes['period'].setEnabled(False)
            self.scan_button.setStyleSheet("background-color: green")
            self.scan_button.setText('Scanning')

    @QtCore.pyqtSlot(float)
    def scan_step(self, val):
        self.pBarEvent(val)
    
    @QtCore.pyqtSlot(float)
    def scan_end(self, val):
        self.signal_to_scan_layout.spinboxes[self.name].setValue(val)
        self.scan_button.setText('Start scan')
        self.scan_button.setStyleSheet('background-color: None')
        self.signal_to_scan_layout.spinboxes[self.name].setEnabled(True)
        self.spinboxes['start'].setEnabled(True)
        self.spinboxes['stop'].setEnabled(True)
        self.spinboxes['period'].setEnabled(True)


    def pBarEvent(self, current_step):
        start = self.spinboxes['start'].value()
        stop = self.spinboxes['stop'].value()    
        self.scanpBar.setValue( 100 * float(current_step-float(start))/float(stop-start)   )
        self.scanLabelValue.setText("%.2f" % current_step + ' V')
        
    def freeze(self):
        self.scan_button.setEnabled(False)        

    def unfreeze(self):
        self.scan_button.setEnabled(True)

if __name__ == '__main__':
    import sys
    from PyQt5.QtWidgets import QMainWindow
    sys.path.append("/home/martindrech/Documents/total_control_app") 
    app = QtGui.QApplication([])
    frame = QtGui.QFrame()
    area = dockarea.DockArea()
    grid = QtGui.QGridLayout()
    frame.setLayout(grid)
    myDock = dockarea.Dock('397', (100, 100))
    area.addDock(myDock)
    grid.addWidget(area, 0, 0)
    l = LaserControlLayout('397', checkboxes = ['cb1', 'cb2', 'cb3'])
    l_scan = LaserScanLayout('397 scan')
    myDock.addWidget(l)
    myDock.addWidget(l_scan)
    frame.show()
    sys.exit(app.exec_())